"""Tests for config.py."""
import os
import shutil
import sys

import pytest
import pytest_check as check

from cliconfig import load_config, make_config, merge, save_config


def test_merge() -> None:
    """Test merge."""
    config1 = {"a.b": 1, "a": {"b": 2, "c": 3}}
    config2 = {"a": {"c": 1}, "a.d": 6}
    config = merge(config1, config2, allow_new_keys=True, priority="flat")
    check.equal(config, {"a": {"b": 1, "c": 1, "d": 6}})
    config = merge(config1, config2, allow_new_keys=True, priority="unflat")
    check.equal(config, {"a": {"b": 2, "c": 1, "d": 6}})
    with pytest.raises(ValueError, match="New parameter 'a.d' found.*"):
        merge(config1, config2, allow_new_keys=False)
    with pytest.raises(ValueError, match="Unknown dot_prior 'UNKONWN'.*"):
        merge(config1, config2, allow_new_keys=True, priority="UNKONWN")


def test_make_config(capsys: pytest.CaptureFixture) -> None:
    """Test make_config."""
    sys_argv = sys.argv.copy()
    sys.argv = [
        "tests/test_make_config.py.py",
        "--config",
        "tests/configs/config1.yaml,tests/configs/config2.yaml",
        "--param2",
        "6",
    ]
    capsys.readouterr()  # Clear stdout and stderr
    config = make_config("tests/configs/default1.yaml", "tests/configs/default2.yaml")
    captured = capsys.readouterr()
    out = captured.out
    expected_config = {
        "param1": 4,
        "param2": 6,
        "param3": 3,
        "letters": {
            "letter1": "f",
            "letter2": "e",
            "letter3": "c",
            "letter4": "d",
        },
    }
    expected_out = (
        "[CONFIG] Merge 2 default configs, "
        "2 additional configs and 1 CLI parameter(s).\n"
    )
    check.equal(config, expected_config)
    check.equal(out, expected_out)

    # No default configs
    config = make_config(allow_new_keys=True)
    expected_config = {
        "param1": 4,
        "param2": 6,
        "letters": {
            "letter1": "f",
            "letter2": "e",
        },
    }
    check.equal(config, expected_config)

    # No additional configs
    sys.argv = [
        "tests/test_make_config.py.py",
    ]
    config = make_config("tests/configs/default1.yaml", "tests/configs/default2.yaml")
    expected_config = {
        "param1": 1,
        "param2": 2,
        "param3": 3,
        "letters": {
            "letter1": "a",
            "letter2": "b",
            "letter3": "c",
            "letter4": "d",
        },
    }
    check.equal(config, expected_config)

    sys.argv = sys_argv.copy()


def test_save_load_config() -> None:
    """Test save_config and load_config."""
    config1 = {"a": 1, "b": {"c": 2}, "d": [2, 3.0], "e": [{"f": 4}]}
    save_config(config1, "tests/tmp/config.yaml")
    check.is_true(os.path.isfile("tests/tmp/config.yaml"))
    config2 = load_config("tests/tmp/config.yaml")
    check.equal(config1, config2)

    # With default configs
    config = {"param1": 3, "param2": 4, "letters": {"letter1": "e"}}
    save_config(config, "tests/tmp/config2.yaml")
    config = load_config(
        "tests/tmp/config2.yaml",
        default_configs=["tests/configs/default1.yaml", "tests/configs/default2.yaml"],
        allow_new_keys=False,
    )
    expected_config = {
        "param1": 3,
        "param2": 4,
        "param3": 3,
        "letters": {
            "letter1": "e",
            "letter2": "b",
            "letter3": "c",
            "letter4": "d",
        },
    }
    check.equal(config, expected_config)
    # Additional keys when allow_new_keys=False
    config = {"param4": 0}
    save_config(config, "tests/tmp/config3.yaml")
    with pytest.raises(ValueError, match="New parameter 'param4' found.*"):
        load_config(
            "tests/tmp/config3.yaml",
            default_configs=[
                "tests/configs/default1.yaml",
                "tests/configs/default2.yaml",
            ],
            allow_new_keys=False,
        )

    shutil.rmtree("tests/tmp")



"""Integration tests for show.py."""

from cliconfig.show import show_config


def main() -> None:
    """Integration for show_config."""
    config = {
        "model": {
            "s1_ae_config": {
                "in_dim": 2,
                "out_dim": 1,
                "layer_channels": [16, 32, 64],
                "conv_per_layer": 1,
                "residual": False,
                "dropout_rate": 0.0,
            },
            "mask_module_dim": [6, 2],
            "glob_module_dims": [2, 8, 2],
            "conv_block_dims": [32, 64, 128],
        },
        "train": {
            "n_epochs": 100,
            "optimizer": {
                "name": "Adam",
                "lr": 0.001,
                "weight_decay": 0.0,
                "warmup_steps": 0,
            },
        },
        "data": {
            "dataset": "mnist",
            "batch_size": 128,
            "num_workers": 6,
        },
    }
    show_config(config)


if __name__ == "__main__":
    main()



"""Functions to show the configurations."""
from typing import Any, Dict


def show_config(config: Dict[str, Any]) -> None:
    """Show the configurations.

    Parameters
    ----------
    config : Dict[str, Any]
        The configuration to show.
    """

    def pretty_print(config: Dict[str, Any], indent: int = 1) -> None:
        """Pretty print the configuration recursively."""
        for key, value in config.items():
            print(f"{'    ' * indent}{key}: ", end="")
            if isinstance(value, dict):
                print()
                pretty_print(value, indent + 1)
            elif isinstance(value, str):
                print(f"'{value}'")
            else:
                print(value)

    print("Config:")
    # printer.pprint(config)
    pretty_print(config)



"""Functions to manipulate configuration as nested dictionaries."""
import os
import sys
from typing import Any, Dict, List, Optional

import yaml
from flatten_dict import flatten, unflatten

from cliconfig.cli_parser import parse


def make_config(*default_configs: str, allow_new_keys: bool = False) -> Dict[str, Any]:
    """Make a config from default configs and CLI arguments.

    Parameters
    ----------
    default_configs : Tuple[str]
        Paths to default configs. They are merged in order and new keys
        are allowed.
    allow_new_keys : bool, optional
        If True, new keys in CLI are allowed in merged config. Otherwise,
        it raises an error. By default False.

    Raises
    ------
    ValueError
        If allow_new_keys is False and CLI has new keys that are not
        in default configs.

    Returns
    -------
    config : Dict[str, Any]
        The merged config.
    """
    config: Dict[str, Any] = {}
    config_paths, config_cli_params = parse(sys.argv)
    print(
        f"[CONFIG] Merge {len(default_configs)} default configs, "
        f"{len(config_paths)} additional configs and "
        f"{len(config_cli_params)} CLI parameter(s)."
    )
    for default_config_path in default_configs:
        with open(default_config_path, "r", encoding="utf-8") as cfg_file:
            default_config = yaml.safe_load(cfg_file)
        config = merge(config, default_config, allow_new_keys=True)

    for config_path in config_paths:
        with open(config_path, "r", encoding="utf-8") as cfg_file:
            additional_config = yaml.safe_load(cfg_file)
        config = merge(config, additional_config, allow_new_keys=allow_new_keys)
    config = merge(config, config_cli_params, allow_new_keys=allow_new_keys)
    return config


def merge(
    config1: Dict[str, Any],
    config2: Dict[str, Any],
    *,
    allow_new_keys: bool = True,
    priority: str = "flat",
) -> Dict[str, Any]:
    """Merge config2 into config1.

    Parameters
    ----------
    config1 : Dict[str, Any]
        The first configuration.
    config2 : Dict[str, Any]
        The second configuration to merge into config1.
    allow_new_keys : bool, optional
        If True, new keys (that are not in config1) are allowed in config2.
        By default True.
    priority : str, optional
        One of 'flat' or 'unflat'.
        If 'flat', keys with dots at the root like `a.b: ...` (flat keys) have
        priority over unflat keys like `'a(): {'b': ...}` when there are conflicts.
        If 'unflat', unflat keys have priority over flat keys when there are conflicts.

    Raises
    ------
    ValueError
        If priority is not one of 'flat', 'unflat' or 'order'.
    ValueError
        If allow_new_keys is False and config2 has new keys that are not in config1.
    """
    if priority in ("flat", "unflat"):
        # Split flat and unflat keys
        config1_flat = dict(filter(lambda x: "." in x[0], config1.items()))
        config1_unflat = dict(filter(lambda x: "." not in x[0], config1.items()))
        config2_flat = dict(filter(lambda x: "." in x[0], config2.items()))
        config2_unflat = dict(filter(lambda x: "." not in x[0], config2.items()))
        # Flatten unflat keys
        config1_unflat_flat = flatten(config1_unflat, reducer="dot")
        config2_unflat_flat = flatten(config2_unflat, reducer="dot")
        if priority == "flat":
            # Flat keys erase the previous unflat keys with the same name
            config1_flat = {**config1_unflat_flat, **config1_flat}
            config2_flat = {**config2_unflat_flat, **config2_flat}
        else:
            # The previous unflat keys erase the flat keys with the same name
            config1_flat = {**config1_flat, **config1_unflat_flat}
            config2_flat = {**config2_flat, **config2_unflat_flat}
    else:
        raise ValueError(
            f"Unknown dot_prior '{priority}'. Should be one of "
            "'flat', 'unflat' or 'order'."
        )
    if not allow_new_keys:
        # Check that there are no new keys in config2
        for key in config2_flat:
            if key not in config1_flat.keys():
                raise ValueError(
                    f"New parameter '{key}' found that is not in the original config."
                )
    # Merge flat configs
    flat_config = {**config1_flat, **config2_flat}
    # Unflats config
    config = unflatten(flat_config, splitter="dot")
    return config


def save_config(config: Dict[str, Any], path: str) -> None:
    """Save config to a file.

    Parameters
    ----------
    config : Dict[str, Any]
        The configuration to save.
    path : str
        The path to the file to save the configuration.
    """
    dir_path = os.path.dirname(path)
    os.makedirs(dir_path, exist_ok=True)
    with open(path, "w", encoding="utf-8") as cfg_file:
        yaml.dump(config, cfg_file, default_flow_style=False)


def load_config(
    path: str,
    default_configs: Optional[List[str]] = None,
    *,
    allow_new_keys: bool = True,
) -> Dict[str, Any]:
    """Load config from a file and merge into optional default configs.

    Parameters
    ----------
    path : str
        The path to the file to load the configuration.
    default_configs : Optional[List[str]], optional
        Paths to default configs. They are merged in order, new keys are allowed.
        Then, the loaded config is merged into the result. None for no default configs.
        By default None.
    allow_new_keys : bool, optional
        If True, the new keys in the loaded config are allowed in the merged config.
        Otherwise, it raises an error. By default True.

    Raises
    ------
    ValueError
        If allow_new_keys is False and the loaded config has new keys that are not
        in default configs.

    Returns
    -------
    config : Dict[str, Any]
        The config merged from default configs and the loaded config.
    """
    config: Dict[str, Any] = {}
    if default_configs:
        for config_path in default_configs:
            with open(config_path, "r", encoding="utf-8") as cfg_file:
                default_config = yaml.safe_load(cfg_file)
            config = merge(config, default_config, allow_new_keys=True)
    with open(path, "r", encoding="utf-8") as cfg_file:
        loaded_config = yaml.safe_load(cfg_file)
    config = merge(config, loaded_config, allow_new_keys=allow_new_keys)
    return config
